// Copyright (C) 2015 The Qt Company Ltd.
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only

#ifndef QTWEBVIEWFUNCTIONS_H
#define QTWEBVIEWFUNCTIONS_H

#include <QtWebView/qwebview_global.h>

QT_BEGIN_NAMESPACE

namespace QtWebView
{
    Q_WEBVIEW_EXPORT void initialize();
}

QT_END_NAMESPACE

#endif // QTWEBVIEWFUNCTIONS_H
