/* This file was generated by syncqt. */
#ifndef QT_QTWEBVIEWQUICK_VERSION_H
#define QT_QTWEBVIEWQUICK_VERSION_H

#define QTWEBVIEWQUICK_VERSION_STR "6.4.2"

#define QTWEBVIEWQUICK_VERSION 0x060402

#endif // QT_QTWEBVIEWQUICK_VERSION_H
